#!/usr/bin/perl
#
use strict;
use warnings;

###############################################################################
#
# MODULE: r.regression.pl
#
# AUTHORS:
#
# Original author:
# Damien O'Grady  17 May 2012
# James Cook University
# Earth and Environmental Science
# 14-88 McGregor Road
# Smithfield, QLD 4870
# Australia
# email: damien.ogrady@my.jcu.edu.au
# 
# PURPOSE: Pixel-wise linear regression between 2 bands across multiple rasters
# COPYRIGHT: (C) 2012 by the Damien O'Grady
#        This program is free software under the GNU General Public
#        License (>=v2). Read the file COPYING that comes with GRASS
#        for details.
###############################################################################


#%Module
#%  description: Carry out linear regression for each pixel across multiple rasters in which X and Y variables are contained in two bands
#%  keywords: raster, regression, linear
#%End
#
#%option
#% key: inputdirectory
#% type: string
#% description: Directory containing gdal-recognised 2-band raster files or one of the files in that directory
#% gisprompt: old_file,file,input
#% required : no
#%end
#
#%option
#% key: outpre
#% type: string
#% description: Text prepended to output rasters xxx.gradient, xxx.intercept, xxx.r2
#% required : yes
#%end
#
#%option
#% key: xsuffix
#% type: string
#% description: Suffix denoting rasters containing X variable
#% answer: .X
#% required : no
#%end
#
#%option
#% key: ysuffix
#% type: string
#% description: Suffix denoting rasters containing Y variable
#% answer: .Y
#% required : no
#%end
#
#%flag
#% key: d
#% description: Whether bands are to be read from files in an external directory, rather than within GRASS
#%end


# Check whether running GRASS

if ( !$ENV{'GISBASE'} ) {
	    printf(STDERR  "You must be in GRASS GIS to run this program.\n");
	        exit 1;
}
 
push @ARGV,'' unless (@ARGV) > 0;
if( $ARGV[0] ne '@ARGS_PARSED@' ){
	my $arg = "";
	for (my $i=0; $i < @ARGV;$i++) {
		$arg .= " $ARGV[$i] ";
	}
	system("$ENV{GISBASE}/bin/g.parser $0 $arg");
	exit;
}

# Save variables

my $directory=$ENV{GIS_OPT_INPUTDIRECTORY};
my $prefix=$ENV{GIS_OPT_OUTPRE};
my $xsuffix=$ENV{GIS_OPT_XSUFFIX};
my $ysuffix=$ENV{GIS_OPT_YSUFFIX};

chomp $xsuffix; chomp $ysuffix;

if ($ENV{GIS_FLAG_D}) {
	directory_method();
} else {
	internal_method();
}

sub internal_method { # This method uses existing rasters within the current Mapset and is quicker
	check_suffices();
	my @prefices = @{get_eligible_prefices()}; # Check for minimum requirements done within sub-routine
	gradient(\@prefices);
	intercept(\@prefices);
	coeff(\@prefices);
}

sub coeff { # Coefficient of determination
	my @prefices = @{shift(@_)};
	my $sumx = join(' + ',map("float($_$xsuffix)",@prefices));
	my $sumy = join(' + ',map("float($_$ysuffix)",@prefices));
	my $sumxy = join(' + ',map("(float($_$xsuffix) * float($_$ysuffix))",@prefices));
	my $sumx2 = join(' + ',map("pow(float($_$xsuffix),2)",@prefices));
	my $sumy2 = join(' + ',map("pow(float($_$ysuffix),2)",@prefices));
	my $n = join(' + ',map("if(isnull($_$xsuffix)||isnull($_$ysuffix),0,1)",@prefices));
	my $expression = <<"EOF";
r.mapcalc '$prefix.R2 = eval( \\
	SumX = $sumx, \\
	SumY = $sumy, \\
	SumXY = $sumxy, \\
	SumX2 = $sumx2, \\
	SumY2 = $sumy2, \\
	N = $n, \\
	numerator = (N * SumXY) - (SumX * SumY), \\
	denominator = sqrt((N * SumX2) - pow(SumX, 2)) * sqrt((N * SumY2) - pow(SumY, 2)), \\
	R = numerator / denominator, \\
	pow(R,2))'
EOF
	print "Creating Coefficient of Determination raster...\n\n";
	system $expression;
}

sub intercept { 
	my @prefices = @{shift(@_)};
	my $sumx = join(' + ',map("float($_$xsuffix)",@prefices));
	my $sumy = join(' + ',map("float($_$ysuffix)",@prefices));
	my $n = join(' + ',map("if(isnull($_$xsuffix)||isnull($_$ysuffix),0,1)",@prefices));
	my $expression = <<"EOF";
r.mapcalc '$prefix.intercept = eval( \\
	SumX = $sumx, \\
	SumY = $sumy, \\
	N = $n, \\
	(SumY - ($prefix.gradient * SumX))/N)'
EOF
	print "Creating intercept raster...\n\n";
	system $expression;
}

sub gradient {
	my @prefices = @{shift(@_)};
	my $sumx = join(' + ',map("float($_$xsuffix)",@prefices));
	my $sumy = join(' + ',map("float($_$ysuffix)",@prefices));
	my $sumxy = join(' + ',map("(float($_$xsuffix) * float($_$ysuffix))",@prefices));
	my $sumx2 = join(' + ',map("pow(float($_$xsuffix),2)",@prefices));
	my $n = join(' + ',map("if(isnull($_$xsuffix)||isnull($_$ysuffix),0,1)",@prefices));
	my $expression = <<"EOF";
r.mapcalc '$prefix.gradient = eval( \\
	SumX = $sumx, \\
	SumY = $sumy, \\
	SumXY = $sumxy, \\
	SumX2 = $sumx2, \\
	N = $n, \\
	numerator = (N * SumXY) - (SumX * SumY), \\
	denominator = (N * SumX2) - pow(SumX, 2), \\
	numerator / denominator)'
EOF
	print "Creating gradient raster...\n\n";
	system $expression;
}

sub get_eligible_prefices {
	my %xs;
	my @prefices;
	my @x = `g.mlist type=rast pat='*$xsuffix'`;
	chomp @x;
	foreach my $X (@x) {
		my $pre = substr($X,0,length($X)-length($xsuffix));
		$xs{$pre} = 1;
	}
	my @y = `g.mlist type=rast pat='*$ysuffix'`;
	chomp @y;
	foreach my $Y (@y) {
		my $pre = substr($Y,0,length($Y)-length($ysuffix));
		push @prefices, $pre if defined $xs{$pre};
	}
	die("You don't have 2 or more matching pairs of rasters with your specified suffices\n") unless $#prefices > 0;
	print "The following prefices correspond to eligible raster pairs: ".join(', ',@prefices)."\n\n";
	return \@prefices;
}

sub check_suffices {
	my $message =<<'EOF';
	You have not used the '-d' flag, which implies that you choose to use the method whichs seeks rasters from within your current Mapset.

	However, you have not indicated a suffix for both the independent (x) and dependent (y) variables.  These are both required for this method.

	Please see the manual entry for details.
EOF
	die($message) if ((length($xsuffix) < 1) || (length($ysuffix) < 1));
	print "x is '$xsuffix' and y is '$ysuffix'\n\n";
}

sub directory_method { # This method brings files in one-by-one from an external directory, adding to aggregates as it goes.  Much slower but memory is never an issue

	clean_directory();

	# Check existence of directory
	unless (-d $directory) {
		print "The directory $directory does not appear to exist!\n";
		exit;
	}

	# Check for the exitence of files
	my $result = `find $directory -type f`;
	unless (chomp $result) {
		print "There appear to be no files in $directory\n";
		exit;
	}

	system <<'EOF';
	r.mapcalc 'SumX=0'
	r.mapcalc 'SumY=0'
	r.mapcalc 'SumXY=0'
	r.mapcalc 'SumX2=0'
	r.mapcalc 'SumY2=0'
	r.mapcalc 'N=0'
EOF

	my @files = `find $directory -type f`;
	chomp @files;

	foreach my $file (@files) {
		print "Reading $file...\n\n";
		system "r.in.gdal inp=$file out=temp --overwrite";
		my @bands = `g.mlist type=rast pat='temp.?'`; chomp @bands;
		if ($#bands < 1) {
			print "The file $file appears not to contain 2 bands\n";
			cleanup();
			exit;
		}
		system <<'EOF';
		r.mapcalc 'SumX = if(isnull(temp.1)||isnull(temp.2),SumX,if(isnull(SumX),temp.1,SumX + float(temp.1)))'
		r.mapcalc 'SumY = if(isnull(temp.1)||isnull(temp.2),SumY,if(isnull(SumY),temp.2,SumY + float(temp.2)))'
		r.mapcalc 'SumXY = if(isnull(temp.1)||isnull(temp.2),SumXY,(float(temp.1) * float(temp.2)) + SumXY)'
		r.mapcalc 'SumX2 = if(isnull(temp.1)||isnull(temp.2),SumX2,pow(float(temp.1),2) + SumX2)'
		r.mapcalc 'SumY2 = if(isnull(temp.1)||isnull(temp.2),SumY2,pow(float(temp.2),2) + SumY2)'
		r.mapcalc 'N = if(isnull(temp.1)||isnull(temp.2),N,N + 1)'
EOF
	}

	system <<"EOF";
	r.mapcalc "$prefix.gradient=eval(numerator = (N * SumXY) - (SumX * SumY), denominator = (N * SumX2) - pow(SumX, 2), numerator / denominator)"
	r.mapcalc "$prefix.intercept = (SumY - ($prefix.gradient * SumX))/N"
	r.mapcalc "$prefix.R2 = eval(numerator = (N * SumXY) - (SumX * SumY), \\
		denominator = sqrt((N * SumX2) - pow(SumX, 2)) * sqrt((N * SumY2) - pow(SumY, 2)), R = numerator / denominator, pow(R,2))"
EOF

	print "The raster maps $prefix.gradient, $prefix.intercept and $prefix.R2 have been created\n\n";

	cleanup();
}

sub cleanup {
	system 'g.remove --q rast=SumX,SumY,SumXY,SumX2,SumY2,N,temp.1,temp.2';
}

sub clean_directory { # Just in case a file has been selected, rather than the actual directory
	unless (-d $directory) {
		until (($directory =~ /\/$/) || (length($directory)==0)) {
			$directory = substr($directory, 0, length($directory)-1);
		}
	}
}
